@@ -26,6 +26,7 @@ from __future__ import print_function
 # need a cleanup. It could be completely reengineered as well.
 
 import contextlib
+import collections
 import copy
 import optparse
 import os
@@ -97,6 +98,8 @@ def find_pylintrc():
             curdir = os.path.abspath(os.path.join(curdir, '..'))
             if os.path.isfile(os.path.join(curdir, 'pylintrc')):
                 return os.path.join(curdir, 'pylintrc')
+            if os.path.isfile(os.path.join(curdir, '.pylintrc')):
+                return os.path.join(curdir, '.pylintrc')
     if 'PYLINTRC' in os.environ and os.path.exists(os.environ['PYLINTRC']):
         pylintrc = os.environ['PYLINTRC']
     else:
@@ -132,15 +135,7 @@ class UnsupportedAction(Exception):
     """raised by set_option when it doesn't know what to do for an action"""
 
 
-def _choice_validator(optdict, name, value):
-    if value not in optdict['choices']:
-        msg = "option %s: invalid value: %r, should be in %s"
-        raise optparse.OptionValueError(msg % (name, value, optdict['choices']))
-    return value
-
-
-def _multiple_choice_validator(optdict, name, value):
-    choices = optdict['choices']
+def _multiple_choice_validator(choices, name, value):
     values = utils._check_csv(value)
     for value in values:
         if value not in choices:
@@ -149,6 +144,12 @@ def _multiple_choice_validator(optdict, name, value):
     return values
 
 
+def _choice_validator(choices, name, value):
+    if value not in choices:
+        msg = "option %s: invalid value: %r, should be in %s"
+        raise optparse.OptionValueError(msg % (name, value, choices))
+    return value
+
 # pylint: disable=unused-argument
 def _csv_validator(_, name, value):
     return utils._check_csv(value)
@@ -178,8 +179,9 @@ VALIDATORS = {
     'regexp': re.compile,
     'csv': _csv_validator,
     'yn': _yn_validator,
-    'choice': _choice_validator,
-    'multiple_choice': _multiple_choice_validator,
+    'choice': lambda opt, name, value: _choice_validator(opt['choices'], name, value),
+    'multiple_choice': lambda opt, name, value: _multiple_choice_validator(opt['choices'],
+                                                                           name, value),
 }
 
 def _call_validator(opttype, optdict, option, value):
@@ -247,6 +249,10 @@ def _patch_optparse():
         optparse.HelpFormatter.expand_default = orig_default
 
 
+def _multiple_choices_validating_option(opt, name, value):
+    return _multiple_choice_validator(opt.choices, name, value)
+
+
 class Option(optparse.Option):
     TYPES = optparse.Option.TYPES + ('regexp', 'csv', 'yn', 'multiple_choice')
     ATTRS = optparse.Option.ATTRS + ['hide', 'level']
@@ -254,7 +260,7 @@ class Option(optparse.Option):
     TYPE_CHECKER['regexp'] = _regexp_validator
     TYPE_CHECKER['csv'] = _csv_validator
     TYPE_CHECKER['yn'] = _yn_validator
-    TYPE_CHECKER['multiple_choice'] = _multiple_choice_validator
+    TYPE_CHECKER['multiple_choice'] = _multiple_choices_validating_option
 
     def __init__(self, *opts, **attrs):
         optparse.Option.__init__(self, *opts, **attrs)
@@ -432,7 +438,7 @@ class OptionsManagerMixIn(object):
         # list of registered options providers
         self.options_providers = []
         # dictionary associating option name to checker
-        self._all_options = {}
+        self._all_options = collections.OrderedDict()
         self._short_options = {}
         self._nocallback_options = {}
         self._mygroups = {}
@@ -635,7 +641,10 @@ class OptionsManagerMixIn(object):
 
     def load_configuration(self, **kwargs):
         """override configuration according to given parameters"""
-        for opt, opt_value in kwargs.items():
+        return self.load_configuration_from_config(kwargs)
+
+    def load_configuration_from_config(self, config):
+        for opt, opt_value in config.items():
             opt = opt.replace('_', '-')
             provider = self._all_options[opt]
             provider.set_option(opt, opt_value)
